"""Audio tag handling code."""

import abc
import base64
import calendar
import collections
import datetime
import functools
import io
import itertools
import logging
import operator
import re
import string
from typing import Any, Deque, Dict, List, Optional, Sequence, Tuple

import magic
import more_itertools
import mutagen
import PIL.Image
import unidecode

from amg import sanitize


class TitleNormalizer:
    """
    Class to chain all title tag transformations.

    What was once a small concise function has grown into the largest part of this project.
    This chains many checks to remove crap from title tags, because record people do not have conventions or
    common sense for that matter, and often try to stuff their grocery list and the name of their cat into
    what should be a simple title for the song.
    """

    def __init__(self, artist: str, album: str, record_label: Optional[str] = None):  # noqa: C901
        self.cleaners: List[Tuple[TitleCleanerBase, Tuple[Any]]] = []

        # TODO separate cleaner from params and copy/reuse cleaners in TitleNormalizer.cleanup

        # remove consecutive spaces
        self.registerCleaner(FunctionCleaner(lambda x: " ".join(x.split()), execute_once=True))

        # detect and remove '#hashtag' suffix
        self.registerCleaner(RegexSuffixCleaner(r"(#[\w]+ ?)+", contains=("#"), execute_once=True))

        # detect and remove 'taken from album xxx, out (on) yyy' suffix
        self.registerCleaner(
            RegexSuffixCleaner("taken from .+, out ", contains=("taken from", "out"), execute_once=True)
        )

        # detect and remove 'a track of the upcoming xxx' suffix
        self.registerCleaner(
            RegexSuffixCleaner("a track of upcoming ", contains=("a track of upcoming"), execute_once=True)
        )

        # detect and remove 'episode x/y of' suffix
        self.registerCleaner(RegexSuffixCleaner("episode [0-9/]+( of)", contains=("episode"), execute_once=True))

        # detect and remove 'album: xxx track yy'
        self.registerCleaner(RegexCleaner("(album: .+ )?track [0-9]+", contains=("track",), execute_once=True))

        # detect and remove 'from xxx LP' suffix
        self.registerCleaner(RegexSuffixCleaner("(taken )?from .+ LP", contains=("from", "LP"), execute_once=True))

        # detect and remove 'from xxx album' suffix
        self.registerCleaner(RegexSuffixCleaner("(taken )?from .*album", contains=("from", "album"), execute_once=True))

        # detect and remove 'xxx out: yy.zz.aa' suffix
        self.registerCleaner(
            RegexSuffixCleaner(r" [\[\(]?([^ ]+ out: )?[0-9]+\.[0-9]+\.[0-9]+[\]\)]?", execute_once=True)
        )

        # detect and remove 'out yy.zz' suffix
        self.registerCleaner(RegexSuffixCleaner(" out [0-9]+/[0-9]+", contains=(" out ",), execute_once=True))

        # detect and remove 'out month xxth' suffix
        self.registerCleaner(RegexSuffixCleaner(" out [a-z]+ [0-9]+th", contains=(" out ",), execute_once=True))

        # detect and remove 'new album out xxx' suffix
        self.registerCleaner(RegexSuffixCleaner("new album out .*$", contains=("new album out ",), execute_once=True))

        # detect and remove '[xxx music]' suffix
        self.registerCleaner(RegexSuffixCleaner(r"[\[\( ][a-z]+ music$", suffixes=("music",), execute_once=True))

        # detect and remove 'xxx entertainment' suffix
        self.registerCleaner(
            RegexSuffixCleaner(r"[\[\( ][a-z]+ entertainment$", suffixes=("entertainment",), execute_once=True)
        )

        # detect and remove 'record label xxx' suffix
        self.registerCleaner(
            RegexSuffixCleaner("record label:? [a-z0-9 ]+$", contains=("record label",), execute_once=True)
        )

        # detect and remove 'next concert: xxx' suffix
        self.registerCleaner(RegexSuffixCleaner("next concert: .+$", contains=("next concert: ",), execute_once=True))

        # detect and remove 'feat.xxx' suffix
        self.registerCleaner(RegexSuffixCleaner(r"feat\..+$", contains=("feat.",), execute_once=True))

        # detect and remove 'ft. xxx'
        self.registerCleaner(RegexCleaner(r"[\(\[ ]+ft\. [a-zA-Z\.\: ]+[\)\]]?", contains=("ft.",), execute_once=True))

        # detect and remove '(xxx productions)'
        self.registerCleaner(
            RegexCleaner(r"[^\w\s].+ productions?[^\w\s]", contains=("production",), execute_once=True)
        )

        # detect and remove 'xxx productions' prefix
        self.registerCleaner(RegexPrefixCleaner(r"^[\w\s]+ productions?", contains=("production",), execute_once=True))

        # detect and remove record label suffix
        if record_label is not None:
            self.registerCleaner(SimpleSuffixCleaner(), (record_label,))

        # detect and remove '- xxx metal' suffix
        base_genres = [
            "crust",
            "black",
            "death",
            "doom",
            "grind",
            "grindcore",
            "progressive",
            "sci-fi",
            "thrash",
        ]
        composed_genres = tuple(
            genre_sep.join(pair) for pair in itertools.permutations(base_genres, 2) for genre_sep in "/- "
        )
        metal_genres = tuple(f"{genre} metal" for genre in tuple(base_genres) + composed_genres)
        base_genres.append("metal")
        for too_common_word in ("black", "death", "thrash"):
            base_genres.remove(too_common_word)
        for genre in metal_genres + composed_genres + tuple(base_genres):
            self.registerCleaner(
                RegexSuffixCleaner(
                    r"[|\(\[/\] -]+(?:[0-9a-z/-\\,]+[ ]*)*" + genre + "( song)?$",
                    suffixes=(genre, f"{genre} song"),
                    execute_once=True,
                    remove_if_skipped=False,
                )
            )

        # detect and remove '(thrash/death from whatever)' suffix
        for genre in metal_genres + composed_genres + tuple(base_genres):
            self.registerCleaner(
                RegexSuffixCleaner(
                    r"[|\(\[/]+[ ]*" + genre + r" from [a-zA-Z-, ]+[\)\]]?$",
                    contains=(f"{genre} from ",),
                    execute_once=True,
                )
            )

        # detect and remove 'xxx metal' prefix
        for genre in metal_genres + composed_genres:
            self.registerCleaner(SimplePrefixCleaner(execute_once=True), (genre,))

        # detect and remove 'xxx productions' suffix
        self.registerCleaner(RegexSuffixCleaner(r"[\[\( ][a-z ]+ productions$", suffixes=(" productions",)))

        # detect and remove track number prefix
        self.registerCleaner(RegexPrefixCleaner("^[0-9]+[ -.]+"))

        # detect and remove 'xxx records' suffix
        self.registerCleaner(RecordsSuffixCleaner("recordings"))
        self.registerCleaner(RecordsSuffixCleaner("records"))

        # detect and remove ' | xxx' suffixes
        self.registerCleaner(RegexSuffixCleaner(r" \| .*$", suffixes=" | ", execute_once=True))

        # build list of common useless expressions
        expressions = set()
        words1 = ("", "explicit", "full", "including", "new", "official", "stop motion", "the new")
        words2 = (
            "",
            "360",
            "album",
            "album/tour",
            "audio",
            "game",
            "lyric",
            "lyrics",
            "music",
            "promo",
            "single",
            "song",
            "stream",
            "studio",
            "track",
            "video",
            "visual",
        )
        words3 = (
            "4k",
            "album",
            "audio",
            "clip",
            "discovery",
            "edit",
            "excerpt",
            "hq",
            "in 4k",
            "lyric",
            "lyrics",
            "only",
            "premier",
            "premiere",
            "presentation",
            "promo",
            "single",
            "song",
            "stream",
            "streaming",
            "teaser",
            "track",
            "trailer",
            "version",
            "video",
            "visualiser",
            "visualizer",
            "vr",
        )
        for w1 in words1:
            for w2 in words2:
                for w3 in words3:
                    if w3 != w2:
                        if w1 or w2:
                            for rsep in (" ", "-", ""):
                                rpart = rsep.join((w2, w3)).strip()
                                expressions.add(" ".join((w1, rpart)).strip())
                        else:
                            expressions.add(w3)
        expressions.update(
            (
                "full ep",
                "full-length",
                "hd",
                "official",
                "offical",
                "pre-listening",
                "pre-order now",
                "pre-orders available",
                "prelistening",
                "preorders available",
                "s/t",
                "sw exclusive",
                "trailer for the upcoming album",
                "transcending obscurity",
                "transcending obscurity india",
                "trollzorn",
                "uncensored",
            )
        )
        year = datetime.datetime.today().year
        for y in range(2016, year + 1):
            expressions.add(str(y))
            for month_name, month_abbr in zip(MONTH_NAMES, MONTH_NAMES_ABBR):
                expressions.add(f"{month_name} {y}")
                expressions.add(f"{month_abbr} {y}")
        expressions_list = list(sorted(expressions, key=len, reverse=True))
        expressions_list.remove("song")
        suffix_cleaner = SimpleSuffixCleaner()
        for expression in expressions_list:
            self.registerCleaner(suffix_cleaner, (expression,))
        prefix_cleaner = SimplePrefixCleaner()
        for expression in expressions_list:
            self.registerCleaner(prefix_cleaner, (expression,))

        # detect and remove artist prefix ot suffix
        self.registerCleaner(ArtistCleaner(), (artist,))

        # detect and remove starting parenthesis expression
        self.registerCleaner(StartParenthesesCleaner(execute_once=True))

        # detect and remove album prefix or suffix
        self.registerCleaner(AlbumCleaner(execute_once=True), (album,))

        # fix paired chars
        self.registerCleaner(PairedCharCleaner(execute_once=True))

        # remove some punctuation
        self.registerCleaner(FunctionCleaner(lambda x: x.strip("-"), execute_once=True))

        # normalize case
        self.registerCleaner(FunctionCleaner(sanitize.normalize_tag_case, execute_once=True))

        # post normalize case fix
        self.registerCleaner(FunctionCleaner(lambda x: x.replace("PT.", "pt."), execute_once=True))

    def registerCleaner(self, cleaner, args=()):
        """Register a new cleaner object."""
        assert isinstance(cleaner, TitleCleanerBase)
        self.cleaners.append((cleaner, args))

    def cleanup(self, title: str) -> str:
        """Do the string cleanup by running all cleaners."""
        cur_title = title
        to_del_indexes: Deque[int] = collections.deque()
        start_index = 0

        while self.cleaners:
            cleaner: TitleCleanerBase
            for i, (cleaner, args) in enumerate(itertools.islice(self.cleaners, start_index, None), start_index):
                remove_cur_cleaner = False
                restart_loop = False

                if cleaner.doSkip(cur_title, *args):
                    if cleaner.remove_if_skipped and not cleaner.doKeep():
                        remove_cur_cleaner = True

                else:
                    new_title = cleaner.cleanup(cur_title, *args)
                    if new_title and (new_title != cur_title):
                        logging.getLogger().debug(
                            f"{cleaner.__class__.__name__} changed title tag: "
                            f"{repr(cur_title)} -> {repr(new_title)}"
                        )
                        # update string and remove this cleaner to avoid calling it several times
                        cur_title = new_title
                        remove_cur_cleaner = not cleaner.doKeep()
                        restart_loop = True

                    elif cleaner.execute_once:
                        remove_cur_cleaner = True
                        # this cleaner did not match and we will remove it, continue from same index
                        if start_index == 0:
                            start_index = i

                if remove_cur_cleaner:
                    to_del_indexes.append(i)
                if restart_loop:
                    start_index = 0
                    break

            else:
                # all cleaners have been called and string did not change title
                break

            while to_del_indexes:
                del self.cleaners[to_del_indexes.pop()]

        if cur_title != title:
            logging.getLogger().info(f"Fixed title tag: {repr(title)} -> {repr(cur_title)}")
        return cur_title


class TitleCleanerBase:
    """Base class for all title cleaner subclasses."""

    RCLEAN_CHARS = "".join(c for c in (string.punctuation + string.whitespace) if c not in "!?)-]")
    LCLEAN_CHARS = "".join(c for c in (string.punctuation + string.whitespace) if c not in "(")

    def __init__(self, *, execute_once: bool = False, remove_if_skipped: Optional[bool] = None):
        self.execute_once = execute_once
        self.remove_if_skipped = remove_if_skipped if (remove_if_skipped is not None) else execute_once

    def doSkip(self, title: str, *args) -> bool:
        """Return True if this cleaner can be skipped for this title string."""
        return False

    def doKeep(self) -> bool:
        """Return True if this cleaner should not be removed even if it matched."""
        return False

    @abc.abstractmethod
    def cleanup(self, title: str, *args: Tuple[Any]) -> str:
        """Cleanup a title string, and return the updated string."""
        pass

    def rclean(self, s: str) -> str:
        """Remove garbage at right of string."""
        r = s.rstrip(self.__class__.RCLEAN_CHARS)
        if r.endswith(" -"):
            r = r[:-2].rstrip(self.__class__.RCLEAN_CHARS)
        return r

    def lclean(self, s: str) -> str:
        """Remove garbage at left of string."""
        r = s.lstrip(self.__class__.LCLEAN_CHARS)
        c = unidecode.unidecode_expect_ascii(r.lstrip(self.__class__.LCLEAN_CHARS)).lstrip(self.__class__.LCLEAN_CHARS)
        if c != r:
            r = c
        return r

    @functools.lru_cache(maxsize=32768)
    def rnorm(self, s: str) -> str:
        """Normalize string unicode chars and remove useless chars from its right."""
        return unidecode.unidecode_expect_ascii(s.rstrip(string.punctuation)).rstrip(string.punctuation).lower()

    @functools.lru_cache(maxsize=32768)
    def lnorm(self, s: str) -> str:
        """Normalize string unicode chars and remove useless chars from its left."""
        return unidecode.unidecode_expect_ascii(s.lstrip(string.punctuation)).lstrip(string.punctuation).lower()

    def startslike(self, s: str, pattern: str, *, sep: Optional[str] = None) -> bool:
        """Return True if start of string s is similar to pattern."""
        s = self.lnorm(s)
        pattern = self.rnorm(pattern)
        cut = s[len(pattern) :]
        return s.startswith(pattern) and ((not sep) or (not cut) or (cut[0] in sep))

    def endslike(self, s: str, pattern: str) -> bool:
        """Return True if end of string s is similar to pattern."""
        return self.rnorm(s).endswith(self.rnorm(pattern))

    def rmsuffix(self, s: str, e: str) -> str:
        """Remove string suffix."""
        return s.rstrip(string.punctuation)[: -len(unidecode.unidecode_expect_ascii(e))]

    def rmprefix(self, s: str, e: str) -> str:
        """Remove string prefix."""
        return s.lstrip(string.punctuation)[len(unidecode.unidecode_expect_ascii(e)) :]


class FunctionCleaner(TitleCleanerBase):
    """Cleaner to apply a function to the title string."""

    def __init__(self, func, **kwargs):
        super().__init__(**kwargs)
        assert callable(func)
        self.func = func

    def cleanup(self, title: str) -> str:  # type: ignore
        """See TitleCleanerBase.cleanup."""
        return self.func(title)


class SimplePrefixCleaner(TitleCleanerBase):
    """Cleaner to remove a static string prefix."""

    def cleanup(self, title: str, prefix: str) -> str:  # type: ignore
        """See TitleCleanerBase.cleanup."""
        if self.startslike(title, prefix, sep=string.punctuation + string.whitespace):
            title = self.lclean(self.rmprefix(title, prefix))
        return title


class SimpleSuffixCleaner(TitleCleanerBase):
    """Cleaner to remove a static string suffix."""

    def cleanup(self, title: str, suffix: str) -> str:  # type: ignore
        """See TitleCleanerBase.cleanup."""
        if self.endslike(title, suffix):
            new_title = self.rclean(self.rmsuffix(title, suffix))
            if new_title.lower() != "the":
                title = new_title
        return title


class ArtistCleaner(SimplePrefixCleaner, SimpleSuffixCleaner):
    """Cleaner to remove artist prefix/suffix."""

    def __init__(self, *args, **kwargs):
        self.prefix_removed = False
        self.suffix_removed = False
        super().__init__(*args, **kwargs)

    def doKeep(self) -> bool:
        """See TitleCleanerBase.doKeep."""
        return not self.suffix_removed

    def cleanup(self, title: str, artist: str) -> str:  # type: ignore
        """See TitleCleanerBase.cleanup."""
        artist_variants = tuple(
            more_itertools.unique_everseen(
                (
                    f"{artist} band",
                    artist,
                    artist.replace(" ", ""),
                    artist.replace("and", "&"),
                    artist.replace("&", "and"),
                    artist.replace(", ", " and "),
                    artist.replace(" and ", ", "),
                    artist.replace("’", ""),
                )
            )
        )
        for s, suffix_only in itertools.zip_longest(("by " + artist,) + artist_variants, (True,), fillvalue=False):
            # detect and remove artist prefix
            if (not suffix_only) and (not self.prefix_removed) and self.startslike(title, s):  # type: ignore
                r = SimplePrefixCleaner.cleanup(self, title, s)  # type: ignore
                self.prefix_removed = True
                return r
            # detect and remove artist suffix
            elif (not self.suffix_removed) and self.endslike(title, s):  # type: ignore
                r = SimpleSuffixCleaner.cleanup(self, title, s)  # type: ignore
                self.suffix_removed = True
                return r
        return title


class AlbumCleaner(SimplePrefixCleaner, SimpleSuffixCleaner):
    """Cleaner to remove album prefix/suffix."""

    def cleanup(self, title: str, album: str) -> str:  # type: ignore
        """See TitleCleanerBase.cleanup."""
        # detect and remove album prefix
        if self.startslike(title, album):
            return SimplePrefixCleaner.cleanup(self, title, album)
        # detect and remove album suffix
        elif self.endslike(title, album):
            title = SimpleSuffixCleaner.cleanup(self, title, album)
            # detect and remove album suffix's prefix
            for suffix in ("taken from", "from the album", "from"):
                if self.endslike(title, suffix):
                    new_title = SimpleSuffixCleaner.cleanup(self, title, suffix)
                    if new_title:
                        title = new_title
                        break
        return title


class RegexCleaner(TitleCleanerBase):
    """Cleaner to remove a regex match."""

    def __init__(self, regex: str, *, contains: Sequence[str] = (), flags: int = re.IGNORECASE, **kwargs):
        super().__init__(**kwargs)
        self.regex = re.compile(regex, flags)
        self.contains = contains

    def doSkip(self, title: str, *args) -> bool:
        """See TitleCleanerBase.doSkip."""
        if self.contains:
            lower_title = title.lower()
            skip = not any(map(lower_title.__contains__, self.contains))
            if skip:
                self.remove_if_skipped = True
            return skip
        return super().doSkip(title, *args)

    def cleanup(self, title: str) -> str:  # type: ignore
        """See TitleCleanerBase.cleanup."""
        rstripped = title.rstrip(string.punctuation)
        match = self.regex.search(rstripped)
        if match:
            title = f"{self.rclean(rstripped[:match.start(0)])} {self.lclean(rstripped[match.end(0):])}".rstrip()
        return title


class RegexSuffixCleaner(RegexCleaner):
    """Cleaner to remove a regex suffix match."""

    def __init__(self, regex: str, *, suffixes: Sequence[str] = (), **kwargs):
        super().__init__(regex, **kwargs)
        self.suffixes = suffixes

    def doSkip(self, title: str, *args):
        """See TitleCleanerBase.doSkip."""
        if self.suffixes:
            return not any(self.endslike(title, suffix) for suffix in self.suffixes)
        return super().doSkip(title, *args)

    def cleanup(self, title: str) -> str:  # type: ignore
        """See TitleCleanerBase.cleanup."""
        match = self.regex.search(title.rstrip(string.punctuation))
        if match:
            title = self.rclean(title[: match.start(0)])
        return title


class RegexPrefixCleaner(RegexCleaner):
    """Cleaner to remove a regex prefix match."""

    def cleanup(self, title: str) -> str:  # type: ignore
        """See TitleCleanerBase.cleanup."""
        match = self.regex.search(title)
        if match:
            title = self.lclean(title[match.end(0) :])
        return title


class RecordsSuffixCleaner(RegexSuffixCleaner, SimpleSuffixCleaner):
    """Cleaner to remove record suffix."""

    def __init__(self, record_word, **kwargs):
        self.record_word = record_word
        super().__init__(r"([|\)\(\[]|on)[0-9a-z,/ ]+" + record_word + "$", suffixes=(record_word,), **kwargs)

    def cleanup(self, title: str) -> str:  # type: ignore
        """See TitleCleanerBase.cleanup."""
        # detect and remove 'xxx records' suffix
        match = self.regex.search(title.rstrip(string.punctuation))
        if match:
            # '(xxx yyy records)' suffix
            title = self.rclean(title[: match.start(0)])
        else:
            title = SimpleSuffixCleaner.cleanup(self, title, self.record_word)
            title = self.rclean(" ".join(title.split()[:-1]))
        return title


class StartParenthesesCleaner(TitleCleanerBase):
    """Cleaner to remove parentheses string prefix."""

    def cleanup(self, title: str) -> str:  # type: ignore
        """See TitleCleanerBase.cleanup."""
        # detect and remove starting parenthesis expression
        closing_pos = title.find(")")
        if title.startswith("(") and (closing_pos != (len(title) - 1)) and (len(title[1:closing_pos]) > 1):
            return self.lclean(title[closing_pos + 1 :])
        return title


class PairedCharCleaner(TitleCleanerBase):
    """Cleaner to fix chars that go by pair."""

    def cleanup(self, title: str) -> str:  # type: ignore
        """See TitleCleanerBase.cleanup."""
        # detect and remove unpaired chars
        char_pairs = ((("(", ")"), False), (('"', '"'), False), (("'", "'"), True))
        c1: str
        c2: str
        for (c1, c2), only_at_edges in char_pairs:
            if only_at_edges:
                if title.endswith(c2) and (c1 not in title[:-1]):
                    title = title[:-1]
                elif title.startswith(c1) and (c2 not in title[1:]):
                    title = title[1:]
            else:
                if c1 != c2:
                    if (title.count(c1) + title.count(c2)) == 1:
                        title = title.translate(str.maketrans("", "", c1 + c2))
                else:
                    if title.count(c1) == 1:
                        title = title.translate(str.maketrans("", "", c1))

        # detect and remove parenthesis at start and end
        if title.startswith("(") and title.endswith(")"):
            title = title[1:-1]

        return title


def normalize_title_tag(title: str, artist: str, album: str, record_label: Optional[str] = None) -> str:
    """Remove useless prefix and suffix from title tag string."""
    normalizer = TitleNormalizer(artist, album, record_label)
    return normalizer.cleanup(title)


def tag(
    track_filepath: str,
    review,
    metadata: Dict[str, str],
    cover_data: Optional[bytes],
    record_label: Optional[str] = None,
):
    """Tag an audio file, return tag dict excluding RG/R128 info and album art."""
    logging.getLogger().info(f"Tagging file {track_filepath!r}")
    mf = mutagen.File(track_filepath, easy=True)

    # sanitize tags
    mf["artist"] = sanitize.normalize_tag_case(review.artist)
    mf["album"] = sanitize.normalize_tag_case(review.album)
    try:
        mf["title"] = normalize_title_tag(metadata["title"], review.artist, review.album, record_label)
    except KeyError:
        mf["title"] = mf["album"]
    try:
        mf["comment"] = metadata["description"]
    except KeyError:
        pass
    tags = dict(mf)

    if cover_data is not None:
        if isinstance(mf, mutagen.mp3.EasyMP3) or isinstance(mf, mutagen.easymp4.EasyMP4):
            # EasyXXX helpers do not allow embedding album art, reopen as normal mutagen file
            mf.save()
            mf = mutagen.File(track_filepath)

        # embed album art
        embed_album_art(mf, cover_data)

    mf.save()

    return tags


def has_embedded_album_art(filepath: str) -> bool:
    """Return True if file already has an embedded album art, False instead."""
    mf = mutagen.File(filepath)
    if isinstance(mf, mutagen.ogg.OggFileType):
        return "metadata_block_picture" in mf
    elif isinstance(mf, mutagen.mp3.MP3):
        return any(map(operator.methodcaller("startswith", "APIC:"), mf.keys()))
    elif isinstance(mf, mutagen.mp4.MP4):
        return "covr" in mf
    return False


def embed_album_art(mf: mutagen.File, cover_data: bytes):
    """Embed album art into audio file."""
    mime = magic.from_buffer(cover_data, mime=True)
    if isinstance(mf, mutagen.ogg.OggFileType):
        picture = mutagen.flac.Picture()
        picture.data = cover_data
        picture.type = mutagen.id3.PictureType.COVER_FRONT
        picture.mime = mime
        encoded_data = base64.b64encode(picture.write())
        mf["metadata_block_picture"] = encoded_data.decode("ascii")
    elif isinstance(mf, mutagen.mp3.MP3):
        mf.tags.add(mutagen.id3.APIC(mime=mime, type=mutagen.id3.PictureType.COVER_FRONT, data=cover_data))
        mf.save()
    elif isinstance(mf, mutagen.mp4.MP4):
        if mime == "image/jpeg":
            fmt = mutagen.mp4.AtomDataType.JPEG
        elif mime == "image/png":
            fmt = mutagen.mp4.AtomDataType.PNG
        else:
            # convert to jpeg
            in_bytes = io.BytesIO(cover_data)
            img = PIL.Image.open(in_bytes)
            out_bytes = io.BytesIO()
            img.save(out_bytes, format="JPEG", quality=85, optimize=True)
            cover_data = out_bytes.getvalue()
            fmt = mutagen.mp4.AtomDataType.JPEG
        mf["covr"] = [mutagen.mp4.MP4Cover(cover_data, imageformat=fmt)]


# copy month names before changing locale
MONTH_NAMES = calendar.month_name[1:]
MONTH_NAMES_ABBR = calendar.month_abbr[1:]



"""String sanitization related tests."""

import unittest

import amg.sanitize as sanitize


class TestSanitize(unittest.TestCase):
    """String sanitization test suite."""

    def test_normalize_tag_case(self):
        """Fix tag capitalization."""
        references = {
            "A Little Test": "A Little Test",
            "I Like L.A": "I Like L.A",
            "Of The Moon": "Of the Moon",
            "Just A Bunch Of Letters": "Just a Bunch of Letters",
            "Episode VI": "Episode VI",
            "EPISODE VIA": "Episode Via",
            "VI VI VI": "VI VI VI",
            "Episode VI: name": "Episode VI: Name",
            "Matsya - The Fish": "Matsya - The Fish",
            "I'M ALIVE!": "I'm Alive!",
            "MARK OF THE BEAST PT. 2: SCION OF DARKNESS": "Mark of the Beast PT. 2: Scion of Darkness",
            "BZZ: THE": "Bzz: The",
            "薄氷(Thin Ice)": "薄氷 (Thin Ice)",
            "III-III: Imha Tarikatı (Sect of Destruction)": "III-III: Imha Tarikatı (Sect of Destruction)",
            "Cosa Del Pantano": "Cosa del Pantano",
            "Lunatic-Liar-Lord": "Lunatic-Liar-Lord",
            "The Day After 'Trinity'": "The Day After 'Trinity'",
            "Amalgamations of Gore/Skin Display": "Amalgamations of Gore/Skin Display",
        }
        for before, after in references.items():
            self.assertEqual(sanitize.normalize_tag_case(before), after)



"""Remove unwanted chars or patterns from strings."""

import itertools
import string
from typing import List, Optional, Tuple

import unidecode

VALID_PATH_CHARS = frozenset(f"-_.()!#$%%&'@^{{}}~ {string.ascii_letters}{string.digits}")
TAG_LOWERCASE_WORDS = frozenset(
    (
        "a",
        "an",
        "and",
        "at",
        "del",
        "for",
        "from",
        "in",
        "of",
        "on",
        "or",
        "over",
        "the",
        "to",
        "with",
        "de",
        "des",
        "du",
        "le",
        "la",
        "les",
        "by",
        "vs",
    )
)


def sanitize_for_path(s: str) -> str:
    """Sanitize a string to be FAT/NTFS friendly when used in file path."""
    s = s.translate(str.maketrans("/\\|*", "---x"))
    s = "".join(c for c in unidecode.unidecode_expect_ascii(s) if c in VALID_PATH_CHARS)
    s = s.strip()
    s = s.rstrip(".")  # this if for FAT on Android
    return s


def normalize_tag_case(s: str) -> str:
    """Normalize case of an audio tag string."""

    def split_sep_char(s: str) -> List[str]:
        words = [s]
        for split_char in "(-":
            new_words = []
            for word in words:
                parts = word.split(split_char, 1)
                if len(parts) > 1:
                    parts[1] = split_char + parts[1]
                    parts = list(filter(None, parts))
                new_words.extend(parts)
            words = new_words
        return words

    def split_words(s: str) -> Tuple[str, ...]:
        return tuple(itertools.chain.from_iterable(split_sep_char(w) for w in s.split()))

    def remove_punct(s: str) -> str:
        return "".join(c for c in s if c not in string.punctuation)

    # case heuristic
    old_words = split_words(s)
    new_words = []
    prev_word: Optional[str] = None
    roman_letters = frozenset("IVXLCDM")
    punct_followed_all_uppercase = set(".-")
    punct_followed_uppercase = set(string.punctuation)
    punct_followed_uppercase.remove("'")
    for i, old_word in enumerate(old_words):
        if (
            (prev_word is not None) and ((prev_word[-1] in punct_followed_all_uppercase) and old_word[0].isupper())
        ) or ("." in old_word):
            new_word = old_word
        elif old_word[0] in "[(-'":
            new_word = old_word
        elif "/" in old_word:
            new_word = old_word
        elif old_word.find("'") == 1:
            if (i > 0) and (old_word[0] != "I"):
                new_word = "'".join((old_word[0].lower(), old_word[2:].capitalize()))
            else:
                new_word = old_word
        elif (
            (i != 0)
            and (old_word.lower() in TAG_LOWERCASE_WORDS)
            and (prev_word is not None)
            and (prev_word[-1] not in punct_followed_uppercase)
        ):
            new_word = old_word.lower()
        elif all(map(roman_letters.__contains__, remove_punct(old_word))):
            new_word = old_word
        else:
            new_word = old_word.capitalize()
        new_word = new_word.replace("I'M", "I'm")
        new_words.append(new_word)
        prev_word = old_word

    # join
    to_join = []
    prev_word = None
    while new_words:
        new_word = new_words.pop(0)
        if (prev_word is None) or ((prev_word is not None) and new_word.startswith("-") and (new_word != "-")):
            to_join.append(new_word)
        else:
            to_join.append(f" {new_word}")
        prev_word = new_word

    return "".join(to_join)



"""Formatter for the logging module, coloring terminal output according to error criticity."""

import enum
import logging
import sys

Colors = enum.Enum("Colors", ("RED", "GREEN", "YELLOW", "BLUE"))

LEVEL_COLOR_MAPPING = {logging.WARNING: Colors.YELLOW, logging.ERROR: Colors.RED, logging.CRITICAL: Colors.RED}
LEVEL_BOLD_MAPPING = {logging.WARNING: False, logging.ERROR: False, logging.CRITICAL: True}


class ColoredFormatter(logging.Formatter):
    """Logging formatter coloring terminal output according to error criticity."""

    def format(self, record):
        """See logging.Formatter.format."""
        message = super().format(record)
        if sys.stderr.isatty() and not sys.platform.startswith("win32"):
            try:
                color_code = LEVEL_COLOR_MAPPING[record.levelno].value
                bold = LEVEL_BOLD_MAPPING[record.levelno]
            except KeyError:
                pass
            else:
                message = f"\033[{bold:d};{30 + color_code}m{message}\033[0m"
        return message



"""Show YoutubeDL download progress with a tqdm progress bar."""

import logging
import os
import shutil

import tqdm


def ltrunc(s: str, length: int) -> str:
    """Truncate string from left."""
    assert length > 0
    if len(s) <= length:
        return s
    return f"…{s[-(length - 1):]}"


class ytdl_tqdm:
    """Convenient context manager to report ytdl download progress."""

    def __init__(self, ytdl_opts=None, **kwargs):
        """See tqdm.tqdm for args description."""
        self.tqdm = None
        self.prev_downloaded_bytes = 0
        self.tqdm_kwargs = kwargs
        if ytdl_opts is not None:
            self.setup_ytdl(ytdl_opts)

    def __bool__(self):
        """Return True if there is an associated progress bar, False instead."""
        return self.tqdm is not None

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self.tqdm is not None:
            return self.tqdm.close()

    def setup_ytdl(self, ytdl_opts):
        """Initialize tqdm bar, update YoutubeDL options, and return them."""
        ytdl_opts.update({"quiet": True, "no_warnings": True, "logger": logging.getLogger()})
        ytdl_opts.setdefault("progress_hooks", []).append(self._log_progress)

        assert self.tqdm is None
        self.tqdm = self._get_new_tqdm()

        return ytdl_opts

    def _log_progress(self, ytdl_state):
        """Report youtube-dl progress (callback)."""
        if ytdl_state["status"] != "downloading":
            return

        # get current state
        downloaded_bytes = ytdl_state["downloaded_bytes"]
        try:
            total_bytes = ytdl_state["total_bytes"]
        except KeyError:
            total_bytes = ytdl_state["total_bytes_estimate"]

        # update state
        if self.prev_downloaded_bytes > downloaded_bytes:
            # new YoutubeDL file, reset progress bar
            try:
                self.tqdm.reset()
            except AttributeError:
                # tqdm < 4.32.0
                self.tqdm.close()
                self.tqdm = self._get_new_tqdm()
            newly_downloaded_bytes = downloaded_bytes
        else:
            newly_downloaded_bytes = downloaded_bytes - self.prev_downloaded_bytes

        # update description
        columns = shutil.get_terminal_size((80, 0))[0]
        filename = os.path.basename(ytdl_state["filename"])
        desc = ltrunc(filename, columns // 5)
        self.tqdm.set_description(desc, refresh=False)

        # update bar
        self.tqdm.total = total_bytes
        self.tqdm.update(newly_downloaded_bytes)
        self.prev_downloaded_bytes = downloaded_bytes

    def _get_new_tqdm(self):
        """Set up and return a new tqdm instance."""
        # default args
        tqdm_kwargs = {"unit": "B", "unit_scale": True, "unit_divisor": 1024}
        # merge with user args
        tqdm_kwargs.update(self.tqdm_kwargs)
        return tqdm.tqdm(**tqdm_kwargs)



"""Tag related unit tests."""

import base64
import json
import os
import shutil
import tempfile
import unittest
import urllib.parse

import mutagen

import amg


def download(url, filepath):
    """Download a file."""
    cache_dir = os.getenv("TEST_DL_CACHE_DIR")
    if cache_dir is not None:
        os.makedirs(cache_dir, exist_ok=True)
        cache_filepath = os.path.join(cache_dir, os.path.basename(urllib.parse.urlsplit(url).path))
        if os.path.isfile(cache_filepath):
            shutil.copyfile(cache_filepath, filepath)
            return
    data = amg.fetch_ressource(url)
    with open(filepath, "wb") as f:
        f.write(data)
    if cache_dir is not None:
        shutil.copyfile(filepath, cache_filepath)


class TestTag(unittest.TestCase):
    """Tag related test suite."""

    @classmethod
    def setUpClass(cls):
        """Set up test suite stuff."""
        cls.ref_temp_dir = tempfile.TemporaryDirectory()
        vorbis_filepath = os.path.join(cls.ref_temp_dir.name, "f.ogg")
        download("https://upload.wikimedia.org/wikipedia/en/0/09/Opeth_-_Deliverance.ogg", vorbis_filepath)
        opus_filepath = os.path.join(cls.ref_temp_dir.name, "f.opus")
        download("https://www.dropbox.com/s/xlp1goezxovlgl4/ehren-paper_lights-64.opus?dl=1", opus_filepath)
        mp3_filepath = os.path.join(cls.ref_temp_dir.name, "f.mp3")
        download("https://www.dropbox.com/s/mtac0y8azs5hqxo/Shuffle%2520for%2520K.M.mp3?dl=1", mp3_filepath)
        m4a_filepath = os.path.join(cls.ref_temp_dir.name, "f.m4a")
        download("https://auphonic.com/media/audio-examples/01.auphonic-demo-unprocessed.m4a", m4a_filepath)

    @classmethod
    def tearDownClass(cls):
        """Clean up test suite stuff."""
        cls.ref_temp_dir.cleanup()

    def setUp(self):
        """Set up test case stuff."""
        self.temp_dir = tempfile.TemporaryDirectory()
        for src_filename in os.listdir(__class__.ref_temp_dir.name):
            shutil.copy(os.path.join(__class__.ref_temp_dir.name, src_filename), self.temp_dir.name)
        self.vorbis_filepath = os.path.join(self.temp_dir.name, "f.ogg")
        self.opus_filepath = os.path.join(self.temp_dir.name, "f.opus")
        self.mp3_filepath = os.path.join(self.temp_dir.name, "f.mp3")
        mf = mutagen.File(self.mp3_filepath)
        mf.tags.delall("APIC")
        mf.save()
        self.m4a_filepath = os.path.join(self.temp_dir.name, "f.m4a")
        mf = mutagen.File(self.m4a_filepath)
        del mf["covr"]
        mf.save()

    def tearDown(self):
        """Clean up test case stuff."""
        self.temp_dir.cleanup()

    def test_normalize_title_tag(self):
        """Test title tag normalization."""
        json_filepath = os.path.join(os.path.dirname(__file__), "normalize_title_tag.json")
        with open(json_filepath, "rt") as json_file:
            for test_data in json.load(json_file):
                source = test_data["source"]
                artist = test_data["artist"]
                album = test_data["album"]
                record_label = test_data.get("record_label")
                expected_result = test_data["result"]
                with self.subTest(source=source, expected_result=expected_result, artist=artist, album=album):
                    self.assertEqual(amg.tag.normalize_title_tag(source, artist, album, record_label), expected_result)

    def test_tag(self):
        """Test tagging for various formats."""
        artist = "Artist"
        album = "Album"
        # https://github.com/mathiasbynens/small/blob/master/jpeg.jpg
        cover_data = b"\xff\xd8\xff\xdb\x00C\x00\x03\x02\x02\x02\x02\x02\x03\x02\x02\x02\x03\x03\x03\x03\x04\x06\x04\x04\x04\x04\x04\x08\x06\x06\x05\x06\t\x08\n\n\t\x08\t\t\n\x0c\x0f\x0c\n\x0b\x0e\x0b\t\t\r\x11\r\x0e\x0f\x10\x10\x11\x10\n\x0c\x12\x13\x12\x10\x13\x0f\x10\x10\x10\xff\xc9\x00\x0b\x08\x00\x01\x00\x01\x01\x01\x11\x00\xff\xcc\x00\x06\x00\x10\x10\x05\xff\xda\x00\x08\x01\x01\x00\x00?\x00\xd2\xcf \xff\xd9"
        review = amg.ReviewMetadata(None, artist, album, None, None, None)

        # vorbis
        self.assertFalse(amg.tag.has_embedded_album_art(self.vorbis_filepath))
        amg.tag.tag(self.vorbis_filepath, review, {}, cover_data)
        tags = mutagen.File(self.vorbis_filepath)
        ref_tags = {"artist": [artist], "album": [album]}
        for k, v in ref_tags.items():
            self.assertIn(k, tags)
            self.assertEqual(tags[k], v)
        self.assertIn("metadata_block_picture", tags)
        self.assertEqual(len(tags["metadata_block_picture"]), 1)
        self.assertIn(cover_data, base64.b64decode(tags["metadata_block_picture"][0]))
        self.assertTrue(amg.tag.has_embedded_album_art(self.vorbis_filepath))

        # opus
        self.assertFalse(amg.tag.has_embedded_album_art(self.opus_filepath))
        amg.tag.tag(self.opus_filepath, review, {}, cover_data)
        tags = mutagen.File(self.opus_filepath)
        ref_tags = {"artist": [artist], "album": [album]}
        for k, v in ref_tags.items():
            self.assertIn(k, tags)
            self.assertEqual(tags[k], v)
        self.assertIn("metadata_block_picture", tags)
        self.assertEqual(len(tags["metadata_block_picture"]), 1)
        self.assertIn(cover_data, base64.b64decode(tags["metadata_block_picture"][0]))
        self.assertTrue(amg.tag.has_embedded_album_art(self.opus_filepath))

        # mp3
        self.assertFalse(amg.tag.has_embedded_album_art(self.mp3_filepath))
        amg.tag.tag(self.mp3_filepath, review, {}, cover_data)
        tags = mutagen.File(self.mp3_filepath)
        ref_tags = {"TPE1": [artist], "TALB": [album]}
        for k, v in ref_tags.items():
            self.assertIn(k, tags)
            self.assertEqual(tags[k].text, v)
        self.assertIn("APIC:", tags)
        self.assertIn(cover_data, tags["APIC:"].data)
        self.assertTrue(amg.tag.has_embedded_album_art(self.mp3_filepath))

        # mp4
        self.assertFalse(amg.tag.has_embedded_album_art(self.m4a_filepath))
        amg.tag.tag(self.m4a_filepath, review, {}, cover_data)
        tags = mutagen.File(self.m4a_filepath)
        ref_tags = {"\xa9ART": [artist], "\xa9alb": [album]}
        for k, v in ref_tags.items():
            self.assertIn(k, tags)
            self.assertEqual(tags[k], v)
        self.assertIn("covr", tags)
        self.assertEqual(len(tags["covr"]), 1)
        self.assertEqual(bytes(tags["covr"][0]), cover_data)
        self.assertTrue(amg.tag.has_embedded_album_art(self.m4a_filepath))



"""Terminal menu code to browse tracks."""

import enum
import webbrowser

import cursesmenu

import amg


class AmgMenu(cursesmenu.CursesMenu):
    """Custom menu to choose review/track."""

    UserAction = enum.Enum("UserAction", ("DEFAULT", "OPEN_REVIEW", "DOWNLOAD_AUDIO"))

    def __init__(self, *, reviews, known_reviews, http_cache, mode, selected_idx):
        menu_subtitle = {amg.PlayerMode.MANUAL: "Select a track", amg.PlayerMode.RADIO: "Select track to start from"}
        super().__init__(
            f"AMG Player v{amg.__version__}",
            f"{mode.name.capitalize()} mode: {menu_subtitle[mode]} "
            "(ENTER to play, "
            "D to download audio, "
            "R to open review, "
            "Q to exit)",
            True,
        )
        if selected_idx is not None:
            self.current_option = selected_idx
        review_strings = __class__.reviewsToStrings(reviews, known_reviews, http_cache)
        for index, (review, review_string) in enumerate(zip(reviews, review_strings)):
            self.append_item(ReviewItem(review, review_string, index, self))

    def process_user_input(self):
        """
        Override key handling to add "open review" and "quick exit" features.

        See cursesmenu.CursesMenu.process_user_input
        """
        self.user_action = __class__.UserAction.DEFAULT
        c = super().process_user_input()
        if c in frozenset(map(ord, "rR")):
            self.user_action = __class__.UserAction.OPEN_REVIEW
            self.select()
        elif c in frozenset(map(ord, "dD")):
            # select last item (exit item)
            self.user_action = __class__.UserAction.DOWNLOAD_AUDIO
            self.select()
        elif c in frozenset(map(ord, "qQ")):
            # select last item (exit item)
            self.current_option = len(self.items) - 1
            self.select()

    def get_last_user_action(self):
        """Return last user action when item was selected."""
        return self.user_action

    @staticmethod
    def reviewsToStrings(reviews, known_reviews, http_cache):
        """Generate a list of string representations of reviews."""
        lines = []
        for i, review in enumerate(reviews):
            try:
                play_count = known_reviews.getPlayCount(review.url)
                played = (
                    f"Last played: {known_reviews.getLastPlayed(review.url).strftime('%x %H:%M')} "
                    f"({play_count} time{'s' if play_count > 1 else ''})"
                )
            except KeyError:
                if review.url in http_cache:
                    review_page = amg.fetch_page(review.url, http_cache=http_cache)
                    if amg.get_embedded_track(review_page, http_cache)[0] is None:
                        played = "No track"
                    else:
                        played = "Last played: never"
                else:
                    played = "Last played: never"
            lines.append((f"{review.artist} - {review.album}", played))
        # auto align/justify
        max_lens = [0] * len(lines[0])
        for line in lines:
            for i, s in enumerate(line):
                if len(s) > max_lens[i]:
                    max_lens[i] = len(s)
        sep = "\t"
        for i, line in enumerate(lines):
            lines[i] = "".join(
                (" " if i < 9 else "", sep.join(s.ljust(max_len + 1) for s, max_len in zip(line, max_lens)))
            )
        return lines

    @staticmethod
    def setupAndShow(mode, reviews, known_reviews, http_cache, selected_idx=None):
        """Set up and display interactive menu, return selected review index or None if exist requested."""
        menu = AmgMenu(
            reviews=reviews, known_reviews=known_reviews, http_cache=http_cache, mode=mode, selected_idx=selected_idx
        )
        menu.show()
        idx = menu.selected_option
        return None if (idx == len(reviews)) else (idx, menu.get_last_user_action())


class ReviewItem(cursesmenu.items.SelectionItem):
    """Custom menu item (menu line), overriden to support several actions per item."""

    def __init__(self, review, review_string, index, menu):
        super().__init__(review_string, index, menu)
        self.review = review

    def action(self):
        """React to user action."""
        if self.menu.get_last_user_action() is AmgMenu.UserAction.OPEN_REVIEW:
            webbrowser.open_new_tab(self.review.url)
            self.should_exit = False
        else:
            self.should_exit = True
